/******************************************************************
 * DPUSER - The Next Generation
 *
 * File:     libfits/3d_stuff.cpp
 * Purpose:  Implementation of several 3D data reduction routines
 * Author:   Thomas Ott
 *
 * Revision history
 * ================
 * 1998 - 2002: Implementation
 ******************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_statistics.h>
#include "fits.h"
#include "math_utils.h"
#include "platform.h"

/* Utility functions from the Numerical recipes */
/*-----------------------*/
/* Sort routine Heapsort */
/*-----------------------*/

void sort1(int n, float ra[])
{
    int l,j,ir,i;
    float rra;

    l=(n >> 1)+1;
    ir=n;
    for (;;){
        if (l > 1) rra=ra[--l];
        else {
           rra=ra[ir];
           ra[ir]=ra[1];
           if (--ir ==1) {
              ra[1]=rra;
              return;
           }
        }
        i=l;
        j=l << 1;
        while (j <= ir) {
           if (j < ir && ra[j] < ra[j+1]) ++j;
           if (rra < ra[j]) {
              ra[i] = ra[j];
              j += (i=j);
           }
           else j=ir+1;
        }
        ra[i]=rra;
    }
}

#define NR_END 1
#define FREE_ARG char*

void nrerror(char error_text[])
/* Numerical Recipes standard error handler */
{
	fprintf(stderr,"Numerical Recipes run-time error...\n");
	fprintf(stderr,"%s\n",error_text);
//	fprintf(stderr,"...now exiting to system...\n");
//	exit(1);
}

float *vector(long nl, long nh)
/* allocate a float vector with subscript range v[nl..nh] */
{
	float *v;

	v=(float *)malloc((dpint64) ((nh-nl+1+NR_END)*sizeof(float)));
	if (!v) nrerror("allocation failure in vector()");
	return v-nl+NR_END;
}

void free_vector(float *v, long nl, long nh)
/* free a float vector allocated with vector() */
{
	free((FREE_ARG) (v+nl-NR_END));
}

#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr
void four1(float data[], unsigned long nn, int isign)
{
        unsigned long n,mmax,m,j,istep,i;
        double wtemp,wr,wpr,wpi,wi,theta;
        float tempr,tempi;

        n=nn << 1;
        j=1;
        for (i=1;i<n;i+=2) {
                if (j > i) {
                        SWAP(data[j],data[i]);
                        SWAP(data[j+1],data[i+1]);
                }
                m=n >> 1;
                while (m >= 2 && j > m) {
                        j -= m;
                        m >>= 1;
                }
                j += m;
        }
        mmax=2;
        while (n > mmax) {
                istep=mmax << 1;
                theta=isign*(6.28318530717959/mmax);
                wtemp=sin(0.5*theta);
                wpr = -2.0*wtemp*wtemp;
                wpi=sin(theta);
                wr=1.0;
                wi=0.0;
                for (m=1;m<mmax;m+=2) {
                        for (i=m;i<=n;i+=istep) {
                                j=i+mmax;
                                tempr=wr*data[j]-wi*data[j+1];
                                tempi=wr*data[j+1]+wi*data[j];
                                data[j]=data[i]-tempr;
                                data[j+1]=data[i+1]-tempi;
                                data[i] += tempr;
                                data[i+1] += tempi;
                        }
                        wr=(wtemp=wr)*wpr-wi*wpi+wr;
                        wi=wi*wpr+wtemp*wpi+wi;
                }
                mmax=istep;
        }
}
#undef SWAP

void floatft(float data[], unsigned long n, int isign)
{
        void four1(float data[], unsigned long nn, int isign);
        unsigned long i,i1,i2,i3,i4,np3;
        float c1=0.5,c2,h1r,h1i,h2r,h2i;
        double wr,wi,wpr,wpi,wtemp,theta;

        theta=3.141592653589793/(double) (n>>1);
        if (isign == 1) {
                c2 = -0.5;
                four1(data,n>>1,1);
        } else {
                c2=0.5;
                theta = -theta;
        }
        wtemp=sin(0.5*theta);
        wpr = -2.0*wtemp*wtemp;
        wpi=sin(theta);
        wr=1.0+wpr;
        wi=wpi;
        np3=n+3;
        for (i=2;i<=(n>>2);i++) {
                i4=1+(i3=np3-(i2=1+(i1=i+i-1)));
                h1r=c1*(data[i1]+data[i3]);
                h1i=c1*(data[i2]-data[i4]);
                h2r = -c2*(data[i2]+data[i4]);
                h2i=c2*(data[i1]-data[i3]);
                data[i1]=h1r+wr*h2r-wi*h2i;
                data[i2]=h1i+wr*h2i+wi*h2r;
                data[i3]=h1r-wr*h2r+wi*h2i;
                data[i4] = -h1i+wr*h2i+wi*h2r;
                wr=(wtemp=wr)*wpr-wi*wpi+wr;
                wi=wi*wpr+wtemp*wpi+wi;
        }
        if (isign == 1) {
                data[1] = (h1r=data[1])+data[2];
                data[2] = h1r-data[2];
        } else {
                data[1]=c1*((h1r=data[1])+data[2]);
                data[2]=c1*(h1r-data[2]);
                four1(data,n>>1,-1);
        }
}

void twofft(float data1[], float data2[], float fft1[], float fft2[],
        unsigned long n)
{
        void four1(float data[], unsigned long nn, int isign);
        unsigned long nn3,nn2,jj,j;
        float rep,rem,aip,aim;

        nn3=1+(nn2=2+n+n);
        for (j=1,jj=2;j<=n;j++,jj+=2) {
                fft1[jj-1]=data1[j];
                fft1[jj]=data2[j];
        }
        four1(fft1,n,1);
        fft2[1]=fft1[2];
        fft1[2]=fft2[2]=0.0;
        for (j=3;j<=n+1;j+=2) {
                rep=0.5*(fft1[j]+fft1[nn2-j]);
                rem=0.5*(fft1[j]-fft1[nn2-j]);
                aip=0.5*(fft1[j+1]+fft1[nn3-j]);
                aim=0.5*(fft1[j+1]-fft1[nn3-j]);
                fft1[j]=rep;
                fft1[j+1]=aim;
                fft1[nn2-j]=rep;
                fft1[nn3-j] = -aim;
                fft2[j]=aip;
                fft2[j+1] = -rem;
                fft2[nn2-j]=aip;
                fft2[nn3-j]=rem;
        }
}

void correl(float data1[], float data2[], unsigned long n, float ans[])
{
        unsigned long no2,i;
        float dum,*fft;

        fft=vector(1,n<<1);
        twofft(data1,data2,fft,ans,n);
        no2=n>>1;
        for (i=2;i<=n+2;i+=2) {
                ans[i-1]=(fft[i-1]*(dum=ans[i-1])+fft[i]*ans[i])/no2;
                ans[i]=(fft[i]*dum-fft[i-1]*ans[i])/no2;
        }
        ans[2]=ans[n+1];
        floatft(ans,n,-1);
        free_vector(fft,1,n<<1);
}

void nr_polint(float xa[], float ya[], int n, float x, float *y, float *dy)
{
	int i,m,ns=1;
	float den,dif,dift,ho,hp,w;
	float *c,*d;

	dif=fabs(x-xa[1]);
	c=vector(1,n);
	d=vector(1,n);
	for (i=1;i<=n;i++) {
		if ( (dift=fabs(x-xa[i])) < dif) {
			ns=i;
			dif=dift;
		}
		c[i]=ya[i];
		d[i]=ya[i];
	}
	*y=ya[ns--];
	for (m=1;m<n;m++) {
		for (i=1;i<=n-m;i++) {
			ho=xa[i]-x;
			hp=xa[i+m]-x;
			w=c[i+1]-d[i];
			if ( (den=ho-hp) == 0.0) nrerror("Error in routine polint");
			den=w/den;
			d[i]=hp*den;
			c[i]=ho*den;
		}
		*y += (*dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
	}
	free_vector(d,1,n);
	free_vector(c,1,n);
}



#ifdef WIN
#pragma warning (disable: 4305) // disable warning 'truncation from const double to float'
#endif

float argonlist[][2] = {
{17914.677000, 15.000000},
{17888.677000, 0.350000},
{17823.991000, 1.500000},
{17445.075000, 3.000000},
{17401.908000, 0.220000},
{16940.584000, 50.000000},
{16860.088000, 0.140000},
{16740.078000, 3.000000},
{16549.352000, 2.500000},
{16519.867000, 5.000000},
{16436.575000, 4.000000},
{16264.070000, 0.160000},
{16180.023000, 0.900000},
{16122.656000, 0.120000},
{15989.491000, 4.000000},
{15899.687000, 2.400000},
{15883.164000, 0.400000},
{15816.777000, 0.180000},
{15446.772000, 0.100000},
{15402.640000, 1.200000},
{15353.128000, 0.600000},
{15348.884000, 1.200000},
{15329.344000, 1.500000},
{15301.925000, 5.000000},
{15172.691000, 3.000000},
{15052.567000, 0.120000},
{15046.503000, 7.000000},
{15030.843000, 0.300000},
{14876.973000, 0.110000},
{14785.722000, 0.400000},
{14739.139000, 0.750000},
{0.0, 0.0}
};

float neonlist[][2] = {
{19577.136000, 1.200000},
{20350.238000, 0.800000},
{21041.295000, 3.000000},
{21708.145000, 3.200000},
{22247.345000, 1.200000},
{22428.130000, 1.200000},
{22466.804000, 0.300000},
{22530.404000, 12.000000},
{22661.813000, 1.500000},
{22687.768000, 0.200000},
{23100.514000, 3.000000},
{23260.302000, 6.000000},
{23373.000000, 7.000000},
{23565.362000, 5.000000},
{23636.515000, 24.000000},
{23701.636000, 1.500000},
{23707.611000, 8.500000},
{23709.160000, 8.500000},
{23951.417000, 16.000000},
{23978.115000, 7.000000},
{24098.557000, 1.000000},
{24161.429000, 2.000000},
{24249.638000, 3.000000},
{24365.048000, 11.500000},
{24371.601000, 5.500000},
{24383.359000, 0.800000},
{24447.853000, 3.000000},
{24459.367000, 4.000000},
{24459.678000, 4.000000},
{24776.473000, 1.500000},
{24903.733000, 1.000000},
{24928.870000, 3.000000},
{25161.682000, 1.100000},
{25524.366000, 2.500000},
{0.0, 0.0}
};

bool Fits::merge3d(Fits & data1, int tweak)
{
    int j, ii, m, l, k, ncord, nsort;
    dpint64 i;
	float sum, sum2, a, b, sumc, sumx;
	float mean, stdev;
	float cord[1024], cord1[1024], *isort;
	int pnum, pnumc;
	Fits imagei, imaged, imageds;

	if ((naxis[1] != 256) || (naxis[2] != 256) || (naxis[0] != 2)) return fits_error("Fits::merge3d: must be array of 256x256");
	if ((naxis[1] != data1.naxis[1]) || (naxis[2] != data1.naxis[2])) return fits_error("Fits::merge3d: must be array of same size");
	if (!setType(R4)) return FALSE;
	if (!data1.setType(R4)) return FALSE;
	imagei.copy(data1);
	imaged.copy(data1);
	imageds.copy(data1);
	imaged = 0.0;
	imageds = 0.0;

/*******************************************************************/
/* Now try to eliminate offsets between the two images column by   */ 
/* column if user wants it.                                        */
/*******************************************************************/
      
	if (tweak != 0) {

/* First we eliminate a general offset between the images assuming that  */
/* the background cancellation did not work perfecty well. The offset    */
/* is determined from the first 30 rows. Before we do that, we determine */
/* 'good' pixels. */

		sum  = 0.0;
		sum2 = 0.0;
		pnum = 0;           
        for (i = 0; i < n_elements; i++) {
			if (r4data[i] == 0.0 || imagei.r4data[i] == 0.0) continue;
			imaged.r4data[i] = r4data[i]-imagei.r4data[i];
			if (imaged.r4data[i] == 0.0) continue;
			sum += imaged.r4data[i];
			sum2+= imaged.r4data[i] * imaged.r4data[i];
			pnum++;
		}
		if (pnum <= 1) {
			mean = 0.0;
			stdev = 0.0;
		} else {
			mean=sum / (float)pnum;
			stdev=3.0*sqrt((sum2-sum*mean)/((float)(pnum-1)));  /* the n-1 thingy and 3 sigma*/
		}
		printf("pnum: %i, mean: %f, stdev: %f, sum: %f, sum2: %f\n", pnum, mean, stdev, sum, sum2);
      
		/* exclude everything > 3 sigma */
      
        for (i = 0; i < n_elements; i++) {
			if (fabs(imaged.r4data[i] - mean) > stdev) imaged.r4data[i] = 0.0;
		}
		
    if (tweak / 16 == 1) {    /* First section */
			tweak -= 16;
			
			/* now subtract the general offset */
			sum  = 0.0;
			pnum = 0;           
            for (i = 0; i < n_elements; i++) {
				if (imaged.r4data[i] == 0.0) continue;
				sum += imaged.r4data[i];
				pnum++;
			}
			if (pnum == 0) {
				mean = 0.0;
			} else {
				mean = sum / (float)pnum;
			}
			printf("pnum: %i, mean: %f, sum: %f\n", pnum, mean, sum);
      
			/* Now apply this to imagei */
      
            for (i = 0; i < n_elements; i++) {
				if (imagei.r4data[i] != 0.0) imagei.r4data[i] += mean;
				if (imaged.r4data[i] != 0.0) imaged.r4data[i] -= mean;
			}    
			printf(".tweak1. general offset %f subtracted\n", mean);
		}
		
		if (tweak / 8 == 1) {    /* Second step */			
			tweak -= 8;

			/* Now work in the difference image on each colum separately */
			/* We remove a general tilt from the spectra. Before that,   */
			/* imaged is smoothed columnwise by a running box.           */
       
			m = 31;          /* # Points in the running box, odd number required */
			b = 0.2;         /* weight at both ends of box, compared to 1 in centre */
			a = 4.*(b-1)/(1-m);
			a = a*a;
			b = (m+1.)/2.;                 /* parabola */
      
			for (l = 0; l < naxis[1]; l++) {         /* select a row */
				for (i = 0; i < naxis[2]; i++) {      /* go through 1 row */
					sum=0.0;			  /* data start at 2 & end at 252 */
					pnum=0;
					pnumc=0;
					for (k = i-(m-1)/2; k < i+(m+1)/2; k++) {
						if (k < 2 || k > 251)  continue;         /* data end at 252 */
						if (imaged.r4data[256*l+k] == 0.0)  continue;
						pnum++;
						sum += imaged.r4data[256*l+k];   /**(a*(pnum-b)*(pnum-b)+1);
       pnumc += a*(pnum-b)*(pnum-b)+1; */
					}
					if (pnum != 0) {
						imageds.r4data[256*l+i] = sum / (float)pnum;  /* pnumc*/
					} else { 
						imageds.r4data[256*l+i] = 0.0;
					}
				}
			}
      
			/*  Now determine tilt in each column and remove it such,     */
			/*  that the first rowns are used as references which are not */
			/*  changed.                                                  */
			/*  A free regression fit is:                                 */
			/*  ax+b : a=[<xiyi>-<xi><yi>]/[<xi^2>-<xi>^2]                */
			/*   ->  : a=[xiyi-xi<yi>]/[xi^2-xi<xi>]                      */
			/*         b=<yi>-a<xi>                                       */
			/*   In this specific case xi=  0,1,2,...255                  */
			/*         <xi> = 255/2=127.5    <xi>^2 = 16256.25            */
			/*       <xi^2> = 5559680/256 = 21717.5                       */
			/*   ->    a=[<xiyi>-127.5<yi>]/5461.25                       */
			/*   ->    b=<yi>-a*127.5                                     */
			/*  what still needs to be calculated is <xiyi> and <yi>      */
			/*  Unfortunately blank pixels can have occured in between, so*/
			/*  we have to calculate all parameters                       */
      
			for (i = 0; i < naxis[1]; i++) {      /* 1 column selected */
				sum=0.;          /* yi   */
				sumc=0.;         /* xiyi */
				sumx=0.;         /* xi */
				sum2=0.;         /* xi^2 */
				pnum=0;
				for (k = 0; k < naxis[2]-1; k++) {  
					if (imageds.r4data[i+256*k] == 0.0) continue;
					sum  += imageds.r4data[i+256*k];
					sumc += imageds.r4data[i+256*k]*k;
					sum2 += k*k;
					sumx += k;
					pnum++;
				}
				if (pnum > 2) {  /* hope the division never sucks */
					a = (sumc-sumx*sum/pnum)/(sum2-sumx*sumx/pnum);
					b = (sum-a*sumx)/pnum;
				} else {
					a = 0.0;
					b = 0.0;
				}
            
				/* now correct imagi and imaged */
				/* The function that we add should be 0 at x=14, since we defined */
				/* the mean of the first 30 pixels to be 0 */
 	 
				if (a != 0.0) {
 	  /*  b = -14*a;                          new coefficient */
					for (k=0; k < naxis[2]; k++) {          /* same column */          
					imagei.r4data[i+256*k] += a*k+b;
	    /*   imaged[i+256*k] -= a*k+b;  */
					}
				}        /* else no change */
			}           /* all columns done */
      printf(".tweak2. general column tilt removed\n"); 
		}
		
		if (tweak / 4 == 1) {    /* Third step */
			tweak -= 4;

			/* Now we deal with the constant offset in every column. We assume */
			/* that it is due to redistribution of the flux. So we should      */
			/* divide the offset out. The ratio is derived from the medians of */
			/* the contributions rather than the means.                        */
      
			for (l = 0; l < naxis[1]; l++) {           /* select a column */
				sum  = 0.0;
				sum2 = 0.0;                  /* First selext good pixels */
				pnum = 0;           
				for (k = 0; k < naxis[2]; k++) {   
					if (r4data[l+256*k]==0.0 || imagei.r4data[l+256*k]==0.0) continue;
					imaged.r4data[l+256*k] = r4data[l+256*k]-imagei.r4data[l+256*k];
					sum += imaged.r4data[l+256*k];
					sum2+= imaged.r4data[l+256*k]*imaged.r4data[l+256*k];
					pnum++;
				}
				if (pnum <= 1) {
					continue;
				} else {
					mean=sum / (float)pnum;
					stdev=2.0*sqrt((sum2-sum*mean)/(pnum-1.));  /* the n-1 thingy and 3 sigma*/
				}
         
				/* Do it only if s/N is high enough */

				if (fabs(mean)/stdev < .5) continue;

				/* exclude everything > 3 sigma */
      
				for (k = 0; k < 255; k++) {   /*  */
					if (imaged.r4data[l+256*k] < mean-stdev || imaged.r4data[l+256*k] > mean+stdev) 
					imaged.r4data[l+256*k] = 0.0;               
				}
         
				/* Now deal with the offset */
         
				sum=0.0;			 
				sumx=0.0;			 
				ncord=0;
				for (k = 0; k < naxis[2]; k++) {        /* go through 1 column */
					if (imaged.r4data[l+k*256] == 0.0)  continue;        
					ncord += 1;
					cord[ncord]= r4data[l+k*256];
					cord1[ncord]= imagei.r4data[l+k*256];
				}
				if (ncord > 20) {                 /* Determine median */
					isort = cord1-1;               /* median2 */
					nsort = ncord;
					sort1(nsort,isort);
					if (ncord % 2 == 1) {
						sumx = cord1[ncord/2];
					} else {
						sumx = (cord1[ncord/2]+cord1[ncord/2-1])/2.;
					} 
					if (sumx != 0.) {                /* otherwise no change */
						isort = cord-1;               /* median1 */
						nsort = ncord;
						sort1(nsort,isort);
						if (ncord % 2 == 1) {
							sum = cord[ncord/2];
						} else {
							sum = (cord[ncord/2]+cord[ncord/2-1])/2.;
						} 
						mean = sum/sumx;                 /* This is the factor  */
						if (mean > 0) {                  /* otherwise not usefull */
							for (i = 0; i < naxis[2]; i++) {      /* go through 1 column */
								imagei.r4data[l+i*256] *= mean;
							}
						}
					}
				}     /* else no change */    
			}        /* all columns done */
      
      printf(".tweak3. individual column offset removed\n"); 
		}
		
		if (tweak / 2 == 1) {    /* Fourth step */
			tweak -= 2;
      
			/*  Finally the individual column tilt is removed.            */
			/*  A free regression fit to each individual column works in  */
			/*  the same way as above.                                    */
      
			for (i = 0; i < naxis[1]; i++) {      /* 1 column selected */
				sum  = 0.;
				sum2 = 0.;
				pnum = 0;           
				for (k = 0; k < 255; k++) {   /* First selext good pixels */
					imaged.r4data[i+256*k] = r4data[i+256*k]-imagei.r4data[i+256*k];
					if (imaged.r4data[i+256*k] == 0.0) continue;
					sum += imaged.r4data[i+256*k];
					sum2+= imaged.r4data[i+256*k]*imaged.r4data[i+256*k];
					pnum++;
				}
				if (pnum <= 1) {
					continue;
				} else {
					mean=sum / (float)pnum;
					stdev=1.5*sqrt((sum2-sum*mean)/(pnum-1.));  /* the n-1 thingy and 3 sigma*/
				}
				/* exclude everything > 2 sigma */
      
				for (k = 0; k < 255; k++) {   /*  */
					if (imaged.r4data[i+256*k] < mean-stdev || imaged.r4data[i+256*k] > mean+stdev) 
					imaged.r4data[i+256*k] = 0.0;               
				}

				/* Now determine the tilt */

				sum=0.;          /* yi   */
				sumc=0.;         /* xiyi */
				sumx=0.;         /* xi */
				sum2=0.;         /* xi^2 */
				pnum=0;
				for (k = 0; k < 255; k++) {  
					if (imaged.r4data[i+256*k] == 0.0) continue;
					sum  += imaged.r4data[i+256*k];
					sumc += imaged.r4data[i+256*k]*k;
					sum2 += k*k;
					sumx += k;
					pnum++;
				}
				if (pnum > 2) {  /* hope the division never sucks */
					a = (sumc-sumx*sum/pnum)/(sum2-sumx*sumx/pnum);
					b = (sum-a*sumx)/pnum;
				} else {
					a = 0.0;
					b = 0.0;
				}
            
				/* now correct imagi and imaged */
 	 
				if (a != 0.0) {
					for (k=0; k < 256; k++) {          /* same column */          
						imagei.r4data[i+256*k] += a*k+b;
						imaged.r4data[i+256*k] -= a*k+b;
					}
				}        /* else no change */
			}           /* all columns done */
      
			printf(".tweak4. individual column tilt removed\n"); 
		}
		
		if (tweak == 1) {    /* Fifth step */
      
			/* Final tweak: subtract median from every column */

			for (l = 0; l < naxis[1]; l++) {           /* select a column */
				sum=0;			 
				sumx=0;			 
				ncord=0;
				for (k = 0; k < naxis[2]; k++) {        /* go through 1 column */
					imaged.r4data[l+256*k] = r4data[l+256*k]-imagei.r4data[l+256*k];
					if (imaged.r4data[l+k*256] == 0.0)  continue;        
					cord[ncord]= imaged.r4data[l+k*256];
					ncord++;
				}
				if (ncord > 20) {                 /* Determine median */
					isort = cord-1;              /* median2 */
					nsort = ncord;
					sort1(nsort,isort);
					if (ncord % 2 == 1) {
						sum = cord[ncord/2];
					} else {
						sum = (cord[ncord/2]+cord[ncord/2-1])/2.;
					} 
					for (k = 0; k < naxis[2]; k++) {        /* go through 1 column */
						imagei.r4data[l+k*256] += sum;
					}
				}
			}        /* all columns done */
			printf(".tweak5. final tweak\n"); 

		}        /* end tweak */
		
		
		
		
	}
    naxis[2] = naxis[2] * 2;
	
// make space for second data
	if (!allocateMemory()) return FALSE;
	for (i = naxis[2] - 2, ii = naxis[2] / 2 - 1; i >= 0; ii--, i -= 2) {
		for (j = 0; j < naxis[1]; j++) {
			r4data[C_I(j, i)] = r4data[C_I(j, ii)];
			r4data[C_I(j, i+1)] = imagei.r4data[C_I(j, ii)];
		}
	}
	return(1);
}

// int nint(float x) /* Determine nearest integer to specified float value */
// {
// 	int k;
// 	
// 	k = (int)x;
// 	if ((x - k) <= 0.5 ) {
// 		return k;
// 	}
// 	else {
// 		return k + 1;
// 	}
// }

int megacal3d(Fits & a, const char *obs_run, const char *obs_band)
{
	int i, run, band, linecount, min, max, cur, column, line;
	char filename[256];
	float result[16385], spec[8193], neonw[40], neoni[40], neonline[8193];
	float relcol;
//	float linecen, temp;
	float rcsq, pixcen, a1, a2, a3;
	float sol1, sol2, cenpos, cenpix, wavelag, centreval;
	int cmin, cmax, j, k, n;
	static float maxres;
	static int maxlag;
	static int row, pixnum;
//	FILE *gp;

// check which observing run
	if (!strncmp(obs_run,"lp96",4)) {
		run = 1;
	} else if (!strncmp(obs_run,"ls96",4)) {
		run = 2;
	} else if (!strncmp(obs_run,"ca97",4)) {
		run = 3;
	} else if (!strncmp(obs_run,"a971",4)) {
		run = 4;
	} else if (!strncmp(obs_run,"a972",4)) {
		run = 5;
	} else if (!strncmp(obs_run,"a981",4)) {
		run = 6;
	} else if (!strncmp(obs_run,"a982",4)) {	
		run = 7;
	} else if (!strncmp(obs_run,"a983",4)) {
		run = 8;
	} else if (!strncmp(obs_run,"a984",4)) {
		run = 9;
	} else if (!strncmp(obs_run,"a985",4)) {
		run = 10;
	} else if (!strncmp(obs_run,"a991",4)) {
		run = 11;
	} else if (!strncmp(obs_run,"a992",4)) {
		run = 12;
	} else if (!strncmp(obs_run,"c993",4)) {
		run = 13;
	} else {
		return(-1);
	}

// check band
	if (!strncmp(obs_band,"kn",2)) {
		band = 1;
	} else if (!strncmp(obs_band,"kl",2)) {
	  band = 2;
	} else if (!strncmp(obs_band,"ks",2)) {
	  band = 3;
	} else if (!strncmp(obs_band,"hn",2)) {
	  band = 4;
	} else if (!strncmp(obs_band,"hh",2)) {
	  band = 5;
	} else {
		return(-1);
	}

/* read in the lamp line list from a file, create data in the lamp
array with gaussian lines of width equal to resolution */

/* Changed to read from global variables argonlist, neonlist */

	linecount = 0;
	if ((band == 1) || (band == 2)) {
		sprintf(filename,"neonnew.list");
		while (neonlist[linecount + 1][0] != 0.0) {
			neonw[linecount] = neonlist[linecount][0];
			neoni[linecount] = neonlist[linecount][1];
			linecount++;
		}
	} else if (band == 3) {
		sprintf(filename,"argonk.list");
		while (argonlist[linecount + 1][0] != 0.0) {
			neonw[linecount] = argonlist[linecount][0];
			neoni[linecount] = argonlist[linecount][1];
			linecount++;
		}
	} else {
		sprintf(filename,"argon.list");
		while (argonlist[linecount + 1][0] != 0.0) {
			neonw[linecount] = argonlist[linecount][0];
			neoni[linecount] = argonlist[linecount][1];
			linecount++;
		}
	}
	printf("linelist found in file: %s\n", filename);

/* 	gp = fopen(filename,"r");
	linecount = 0;
	while (fscanf(gp,"%f %f",&linecen,&temp) != EOF) {
		neonw[linecount] = linecen;
		neoni[linecount] = temp;
		linecount++ ;
	}
	fclose(gp); */

	if ((a.Naxis(0) != 2) || (a.Naxis(1) != 256) || (a.Naxis(2) != 512)) {
		printf("megacal3d: must be array of size 256x512\n");
		return FALSE;
	}
	if (!a.setType(R4)) return FALSE;

	for (column=0; column < 256; column++ ) { /* Loop each column in lambda space */
		min = 1;                                       /* prepare for status bar */
		max = 256;
		cur = column+1;

/* initialize arrays */
		for ( i = 1; i <= 8192; i++) {
			neonline[i] = 0;
		}

		relcol = column - 127.5;                    /* Zero in dispersion model 
                                                  corresponds to array centre */

		rcsq = relcol*relcol;
		if ((run == 1) && (band == 1)) {
			pixcen = 2.217840; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4 - 1.327201e-00*relcol + 2.472929e-05*rcsq;
			a2 =   9.698469e+0 - 2.260616e-05*relcol - 1.014594e-07*rcsq;
			a3 =   7.420701e-4 - 8.493901e-08*relcol - 7.714226e-10*rcsq;
		} else if ((run == 2) && (band == 1)) {
			pixcen = 2.172640; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4 - 1.428022e-00*relcol + 1.479699e-03*rcsq;
			a2 =   9.684871e+0 - 4.285905e-05*relcol - 3.230391e-07*rcsq;
			a3 =   7.627958e-4 + 4.973074e-08*relcol + 7.395694e-10*rcsq;
		} else if ((run == 2) && (band == 2)) {
			pixcen = 2.315552; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4 - 8.814002e-01*relcol - 5.425691e-05*rcsq;
			a2 =   5.420413e+0 - 1.906983e-05*relcol - 1.288110e-07*rcsq;
			a3 =   6.838133e-4 - 4.575119e-08*relcol + 3.224833e-10*rcsq;
		} else if ((run == 2) && (band == 3)) {
			pixcen = 2.086464; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4 - 7.370081e-01*relcol + 1.426821e-03*rcsq;
			a2 =   4.830716e+0 - 1.955858e-05*relcol - 1.695582e-07*rcsq;
			a3 =   6.282131e-4 - 2.236271e-08*relcol + 3.444918e-10*rcsq;
		} else if ((run == 2) && (band == 4)) {
			pixcen = 1.632559; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4 - 9.383070e-01*relcol - 7.359458e-05*rcsq;
			a2 =   6.229130e+0 - 3.419027e-05*relcol - 2.167776e-07*rcsq;
			a3 =   5.365792e-4 + 2.556663e-08*relcol + 2.734536e-10*rcsq;
		} else if ((run == 3) && (band == 1)) {
			pixcen = 2.167232; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4 - 1.403047e+00*relcol + 1.449776e-03*rcsq;
			a2 =   9.680334e+00 - 5.377410e-05*relcol - 2.830139e-07*rcsq;
			a3 =   7.490959e-04 - 6.153053e-08*relcol + 7.522528e-10*rcsq;
		} else if ((run == 3) && (band == 5)) {
			pixcen = 1.664933; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4 - 5.165094e-01*relcol + 1.130427e-03*rcsq;
			a2 =   3.764246e+00 - 1.710412e-05*relcol - 1.123260e-07*rcsq;
			a3 =   5.016063e-04 - 1.888443e-08*relcol + 2.233767e-10*rcsq;
		} else if ((run == 4) && (band == 1)) {
			pixcen = 2.199644; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.568798e-00*relcol + 1.353233e-03*rcsq;
			a2 =   9.827351e+00 - 2.841587e-05*relcol + 2.206028e-07*rcsq;
			a3 =   7.296293e-04 + 1.675249e-08*relcol + 2.244198e-09*rcsq;
		} else if ((run == 4) && (band == 4)) {
			pixcen = 1.651617; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 9.969579e-01*relcol + 1.045550e-03*rcsq;
			a2 =   6.314199e+00 - 3.071816e-05*relcol + 5.178511e-08*rcsq;
			a3 =   5.457300e-04 - 3.062932e-09*relcol + 1.254224e-09*rcsq;
		} else if ((run == 5) && (band == 1)) {
			pixcen = 2.198867; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.578211e-00*relcol + 1.476337e-03*rcsq;
			a2 =   9.827213e+00 - 3.402451e-05*relcol + 1.359544e-07*rcsq;
			a3 =   7.297228e-04 + 1.940326e-08*relcol + 1.826487e-09*rcsq;
		} else if ((run == 5) && (band == 4)) {
			pixcen = 1.651207; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.000741e-00*relcol + 1.096481e-03*rcsq;
			a2 =   6.313843e+00 - 2.091752e-05*relcol + 1.193703e-07*rcsq;
		a3 =   5.501770e-04 - 1.711250e-09*relcol + 5.908500e-10*rcsq;
		} else if ((run == 6) && (band == 1)) {
			pixcen = 2.197288; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.566947e-00*relcol + 1.455840e-03*rcsq;
			a2 =   9.812165e+00 - 2.166734e-05*relcol + 1.030817e-07*rcsq;
			a3 =   7.450081e-04 + 4.675547e-08*relcol + 1.152589e-09*rcsq;
		} else if ((run == 6) && (band == 5)) {
			pixcen = 1.676817; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 5.746849e-01*relcol + 1.080278e-03*rcsq;
			a2 =   3.808657e+00 - 8.277254e-06*relcol + 4.357994e-08*rcsq;
			a3 =   5.142785e-04 - 7.621491e-09*relcol + 5.350763e-10*rcsq;
		} else if ((run == 7) && (band == 3)) {
			pixcen = 2.055619; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 8.670402e-01*relcol + 1.587019e-03*rcsq;
			a2 =   4.895298e+00 - 9.486564e-06*relcol + 2.092809e-08*rcsq;
			a3 =   6.422711e-04 - 1.833257e-08*relcol - 2.781284e-11*rcsq;
		} else if ((run == 7) && (band == 2)) {
			pixcen = 2.277080; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.015225e+00*relcol + 1.702074e-03*rcsq;
			a2 =   5.479437e+00 - 4.636515e-06*relcol + 2.341500e-07*rcsq;
			a3 =   7.036939e-04 - 2.239699e-08*relcol + 1.128616e-09*rcsq;
		} else if ((run == 9) && (band == 1)) {
			pixcen = 2.157940; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.675444e+00*relcol + 1.467022e-03*rcsq;
			a2 =   9.808877e+00 - 3.105451e-05*relcol - 4.371717e-09*rcsq;
			a3 =   7.686431e-04 + 2.121690e-08*relcol + 6.202179e-10*rcsq;
		} else if ((run == 9) && (band == 5)) {
			pixcen = 1.660796; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 6.116611e-01*relcol + 1.156054e-03*rcsq;
			a2 =   3.810187e+00 - 1.098161e-05*relcol + 5.248947e-08*rcsq;
			a3 =   5.155353e-04 + 3.125166e-08*relcol + 9.623403e-10*rcsq;
		} else if ((run == 10) && (band == 1)) {
			pixcen = 2.160880; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.659873e+00*relcol + 1.582388e-03*rcsq;
			a2 =   9.810042e+00 - 3.386833e-05*relcol - 1.087019e-07*rcsq;
			a3 =   7.552480e-04 + 4.406794e-08*relcol + 8.541737e-10*rcsq;
		} else if ((run == 10) && (band == 5)) {
			pixcen = 1.661975; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 6.091557e-01*relcol + 1.159916e-03*rcsq;
			a2 =   3.809954e+00 - 1.202910e-05*relcol - 1.406670e-08*rcsq;
			a3 =   5.208868e-04 + 6.621390e-08*relcol + 5.225423e-10*rcsq;
		} else if ((run == 11) && (band == 1)) {
			pixcen = 2.155472; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.639102e+00*relcol + 1.487734e-03*rcsq;
			a2 =   9.813142e+00 - 2.587683e-05*relcol + 1.548016e-07*rcsq;
			a3 =   7.618236e-04 + 6.276021e-08*relcol + 1.461809e-09*rcsq;
		} else if ((run == 12) && (band == 1)) {
			pixcen = 2.161149; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.654354e+00*relcol + 1.505121e-03*rcsq;
			a2 =   9.802176e+00 - 2.595236e-05*relcol + 6.305705e-08*rcsq;
			a3 =   7.684688e-04 + 3.149165e-08*relcol + 2.903716e-10*rcsq;
		} else if ((run == 12) && (band == 5)) {
			pixcen = 1.662267; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 5.773750e-01*relcol + 1.119156e-03*rcsq;
			a2 =   3.811902e+00 - 5.516988e-06*relcol + 8.420231e-08*rcsq;
			a3 =   5.176319e-04 + 8.390235e-09*relcol + 5.320550e-10*rcsq;
		} else if ((run == 13) && (band == 1)) {
			pixcen = 2.142482; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 1.708794e+00*relcol + 1.488836e-03*rcsq;
			a2 =   9.772827e+00 - 3.392730e-05*relcol + 1.278119e-07*rcsq;
			a3 =   7.791834e-04 + 1.727228e-08*relcol + 9.016933e-10*rcsq;
		} else if ((run == 13) && (band == 5)) {
			pixcen = 1.655214; /* initial lambda for pixel 256.5 */
			a1 =   pixcen*1e+4  - 6.290776e-01*relcol + 1.202311e-03*rcsq;
			a2 =   3.788736e+00 - 2.865775e-05*relcol - 1.365472e-07*rcsq;
			a3 =   5.255751e-04 + 5.655139e-08*relcol + 6.384093e-10*rcsq;
 		} else {
			printf("The given set of options is not implemented\n");
			return(-1);
		}

		for (line = 0; line < linecount; line++ ) {       /* Open LINE loop */

                         /* solve for exact offset of line from column centre */

			sol1 = ( a2*(-1.0) + sqrt( a2*a2 - 4*(a1-neonw[line])*a3) )/(2*a3);
			sol2 = ( a2*(-1.0) - sqrt( a2*a2 - 4*(a1-neonw[line])*a3) )/(2*a3);

                                             /* Test for appropriate solution */
			if (    ( sol2 > 256 || sol2 < -256 ) 
			     && ( sol1 < 256 && sol1 > -256 ) ) cenpos = sol1;

			else if (    ( sol1 > 256 || sol1 < -256 ) 
			     && ( sol2 < 256 && sol2 > -256 ) ) cenpos = sol2;

			else {
/*
       sprintf(message, "Column %d, line %d, NO SOLUTION,%f %f", column, line,sol1,sol2);
       anyout_c(&showdev, tofchar(message));
*/
				continue;
			}
  
			/* Magnify image by factor 8 and reset offset zero to bottom of column */
			cenpix = cenpos*8.0 + 2048.0 + 1536.0;   

			/* Determine max and min pixel limits over which line should be convolved */
			cmin = (nint(cenpix)-48) > 1 ? nint(cenpix)-48 : 1 ;
			cmax = (nint(cenpix)+48) < 8192 ? nint(cenpix)+48 : 8192 ;  

                                 /* Convolve neon line with Gaussian function */
			for (j = cmin; j <= cmax; j++) {
				neonline[j] += neoni[line]*exp((double)(-0.5*(j-cenpix)*(j-cenpix)/(7*7))); 
			}                                       /* Close MIN/MAX loop */
	
		}                                                /* Close LINE loop */
     


/* for each column, map the 512 data points onto a 4096 element grid for FFT */

/* first initialize the two arrays each time */
		for (i = 1; i <= 8192; i++) {
			spec[i] = result[i] = 0;
		}
		for (i=8193;i <= 16384; i++) {
			result[i] = 0;
		}

/* now take the data points and put them into the spec array */

		for (j = 0; j < 512; j++) {

/* add offset of 1536 (same as lamp array), scale to improve fft sensitivity */
/* the old routine had a bug here, it was inserting only 1 value
instead of 8 */
			for (k=0; k < 8;k++) {
				if (a.r4data[column+256*j] > 0.0) {
					spec[1536+k+(j*8)] = (a.r4data[column+256*j] ) / 1000.0 ;
				} else {
					spec[1536+k+(j*8)] = 0;
				}
                    /* eliminate negative pixels */
			}
		}

                    /* now call the cross correlation routine */
		n = 8192;
/*
	if ((column >= 1) && (column <= 10)) {
		sprintf(filename,"autocorr%03d",column);
		fp = fopen(filename,"w");
		for (i=4097;i<=8192; i++) {
			fprintf(fp,"%d %e %e %e\n",(i-4096),result[i],neonline[i-4096],spec[i-4096]);
		}
		for (i=1; i<=4096; i++) {
			fprintf(fp,"%d %e %e %e\n",(i+4096),result[i],neonline[i+4096],spec[i+4096]);
		}
		fclose(fp);
	}
*/
		correl(spec,neonline,n,result);

/* only the first 8192 values of result contain float data. find the
maximum of this array. This is the autocorrelation peak */

		maxres = -20000;
		maxlag = -1;
		for (i = 1; i <= 8192; i++) {
			if (maxres < result[i]) {
				maxres = result[i] ;
				maxlag = i;
			}
		}
/*
	fclose(gp);
*/

    /* if maxlag is 1 (zero lag) then pixel 1536 + 2048 is at 
       location a1 in the spec array. Non-zero lags imply that 
       the pixel corresponding to this wavelength is located 
       MAXLAG pixels from the zero point */

    /* Convert FFT pixels to image pixels */
		if (maxlag < 4096) { /* positive lag case */
			wavelag = (1.0-maxlag)/8.0;
		}
		else { /* negative lag case */
			wavelag = (8193.0-maxlag)/8.0;
		}
          /* Determine zero order coefficient (a1) for this column */
  	centreval =   a1 + a2*wavelag - a3*wavelag*wavelag;

/*
		fprintf(gp,"%d %f %f\n",column,wavelag,centreval);
*/
              /* Prepare to write out wavelengths as pixel values */
		for (row = 0; row < 512 ; row++) {
			cenpix = row - 255.5;

			pixnum = column + 256*row;
			a.r4data[pixnum] = (centreval + a2*cenpix + a3*(cenpix*cenpix))/1.0e4;

		}


    
	} /* FOR1 */                  /* End of Column search */  

	return(1);
}

int calibrate3d(Fits & a, const Fits & calset)
{
	float lambda, centerlam, deltalam, imagecol[513], calcol[513];
	float intercept, slope, *calptr, renorm[20], *imageptr, renorm_l;
	float actualval, chisq;
	int k, column, min, max, cur, band, outrow, testrow, ndata, datapos;
	int firstpos;
	Fits outim;
	
	if ((a.Naxis(0) != 2) || (a.Naxis(1) != 256) || (a.Naxis(2) != 512)) {
		printf("calibrate3d: must be array of size 256x512\n");
		return FALSE;
	}
	if ((calset.Naxis(0) != 2) || (calset.Naxis(1) != 256) || (calset.Naxis(2) != 512)) {
		printf("calibrate3d: must be array of size 256x512\n");
		return FALSE;
	}
	outim.create(calset.Naxis(1), 600);

/* we follow the same mnemonic for the band here, as we did in */
/* 3dmegacal so that we are consistent. internal variable band becomes */
/* an int, and the values are as follows: */
/* kn = low resolution K grism = 1 */
/* kl = high resolution long K grism = 2 */
/* ks = high resolution short K grism = 3 */
/* hn = low resolution H band grism = 4 */
/* hh = high resolution H band grism = 5 */

	ndata = 3;

	if (!a.setType(R4)) return FALSE;

	centerlam = calset.ValueAt(256*258+126); /* purposely offset to prevent */
				      /* quadrant borders from */
				      /* affecting the values */
	deltalam = calset.ValueAt(256*500+126) - calset.ValueAt(256*200+126);

	if ((centerlam > 1.95) && (centerlam < 2.15) && (deltalam < 0.22)) {
		printf(">>>>> K-short (high res) calibration <<<<<\n");
		band = 3;
	} else if ((centerlam > 2.1) && (centerlam < 2.3) && (deltalam > 0.22)) {
		printf(">>>>> K-normal (low res) calibration <<<<<\n");
		band = 1;
	} else if ((centerlam > 2.2) && (centerlam < 2.4) && (deltalam < 0.22)) {
		printf(">>>>> K-long (high res) calibration <<<<<\n");
		band = 2;
	} else if ((centerlam > 1.5) && (centerlam < 1.7) && (deltalam > 0.135)) {
		printf(">>>>> H-normal (low res) calibration <<<<<\n");
		band = 4;
	} else if ((centerlam > 1.5) && (centerlam < 1.77) && (deltalam < 0.125)) {
		printf(">>>>> H (high res) calibration <<<<<\n");
		band = 5;
	} else {
		printf(">>>>> Bad calibration file <<<<<\n");
		return(-1);   /* Dummy return */
	}

/* tackle the image a column at a time */
        	   
	for (column=0; column < 256; column++ ) {
		min = 0;
		max = 253;
		cur = column+1;

/* copy the two columns into temporary arrays to speed things up */
		for (k=1; k<513; k++) {
			imagecol[k] = a.r4data[column+256*(k-1)];
			calcol[k] = calset.ValueAt(column+256*(k-1));
		}

/* now loop over the rows */
		for (outrow = 0; outrow < 600; outrow++ ) {

/* establish the lambda for each row */
			if (band == 1) {
				lambda = 1.901 + 0.001*outrow;
			} else if (band == 2) {
				lambda = 2.1505 + 0.0005*outrow;
			} else if (band == 3) {
				lambda = 1.9305 + 0.0005*outrow;
			} else if (band == 4) {
				lambda = 1.4506 + 0.0006*outrow;
			} else if (band == 5) {
				lambda = 1.55037 + 0.00037*outrow;
			} else {
				printf(">>>> Band option not implemented <<<<\n");
				return(-1);   /* Dummy return */
			}   
	   
	     /* this value used to be 1.850. However, I notice that gipsy puts 
	     the first subset of the cubed image at 1.852, so we try to 
	     compensate for this by using 1.852 as our value. In former times, 
	     the wavelengths returned by the cal. routine used to be in error 
	     - NT 18Jun95 */
         /* for the merged routine, the offset is reduced to 0.001, as
         is to be expected. also, we keep with our previous convention
         and choose the first channel to be 1.901 not 1.900. The move
         from 1.85 to 1.90 is in recognition of the fact that we never
         use wavelengths below 1.90 and yet, we do, sometimes, use
         wavelengths somewhat longer than 2.45. The NEW formula for
         wavelength computation is 1.900 + row*0.001 (1 < row < 600)
         NT 25may96 */
		

/* we can only find an answer if lambda lies between the two available
extremes */

			if ((lambda < calcol[1]) || (lambda > calcol[512])) {
				outim.r4data[column+256*outrow] = 0.0;
			}
/* lambda lies in our range */
			else {
				testrow = 1;
				while (lambda > calcol[testrow]) { 
					testrow++;
				}
/* at this point calcol[testrow-1] < lambda <= calcol[testrow] */

/* now datapos is half the number of datapoints. We determine the
start position of the fit array and treat the special cases of being
near the edge */

/* numerical recipes wants unit offset arrays, which means that we
must supply it with an array pointer pointing one location before
the actual data */

				datapos = (int)(ndata+1)/2;
				firstpos = testrow-datapos;

				if (firstpos < 1) {
					firstpos = 1;
				} else if (firstpos > (512-ndata+1)) {
					firstpos = 512 - ndata + 1;
				}
/* we seem to be running into a problem of numerical precision, so
we will rescale the x axis (calcol) before we call svdfit */

				intercept = calcol[firstpos];
				slope = calcol[firstpos+ndata-1]-calcol[firstpos];
				for (k = 1; k < ndata+1; k++) {
					renorm[k] = (calcol[firstpos+k-1] - intercept)/slope;
				}
				calptr = &calcol[firstpos-1];
				imageptr = &imagecol[firstpos-1];
/* try to replace the svdfit routine by a simpler polynomial 
interpolation routine */
				renorm_l = (lambda - intercept)/slope;
				nr_polint(renorm,imageptr,ndata,renorm_l,&actualval,&chisq);
/* now evaluate the polynomial based on the fit coefficients */
/*
	sprintf(message,"%d %f %e %e %e %e",column,lambda,*(imageptr+1),*(imageptr+2),*(imageptr+3),actualval);
	anyoutC(showdev,message);
*/
				outim.r4data[column + outrow*256] = actualval;
			} /* close else */
		} /* Close output row loop */
	}     /* End of loop through all columns -> spatial pixels */
                        /* i.e. End of column search */
	a = outim;
	return(1);
}

//int makeCube3d(Fits & a, char *fname)
//{
// 	int kpix, space, column;
// 	Fits cube;
// 	
// 	cube.setNaxis(1, (int)(sqrt((float)a.Naxis(1))));
// 	cube.setNaxis(2, cube.Naxis(1));
// 	cube.setNaxis(3, a.Naxis(2));
// 	cube.bitpix = T_R4;
// 	cube.io_open(fname, "w+b");
// 	cube.io_wfhdr();
// 	cube.setNaxis(3, 1);
// 	cube = a;
// 	cube = 0.0;
// 	
// 	for (space = 0; space < a.Naxis(2); space++) {
// 		kpix = 4;
// 		for (column = 0; column < 256; column++) {
//             if      (column ==  32) {kpix -=1;} /* Same for the Mar/Aprl 96 */
//             else if (column ==  80) {kpix -=1;} /* KL band. */
//             else if (column == 112) {kpix -=1;}   
//             else if (column == 144) {kpix -=1;}   
//             else if (column == 192) {kpix -=1;}
// 			cube.data[cube.C_I(column, space)] = a.data[kpix-1+space*256];
// 			kpix++;
// 		}
// 	}
// 	cube.io_wdata(0);
// 	fclose(cube.fd);
//	return(1);
//}

//static float dat94[16]={0.00,-0.00,0.40,-0.07, /* adopted from Niranjan */
//                        0.30,-0.07,0.42,-0.00,
//                        0.38,-0.00,0.35,-0.20,
//                        0.10, 0.20,0.35,-0.20};
//
//static float dat95[16]={0.00,0.14,0.36,-.35,  /*Same for 3D configuration */
//                        0.00,0.39,-.20,0.23,   /* Jan 95, K band */
//                        -.34,0.07,0.46,-.19,
//                        0.10,0.33,0.48,-.47};
//
//static float dat95h[16]={0.10,0.24,0.46,-.25,  /*Same for 3D configuration */
//                         0.10,0.49,-.10,0.33,   /* Jan 95, H band */
//                         -.24,0.17,0.56,-0.09,
//                         0.20,0.43,0.58,-0.37};

static float dat96[16]={0.07,0.23,-0.37,-.05,  /*Same for 3D configuration */
                        0.20,-0.19,.09,-0.31,   /* Dec 95/Jan96, K band */ 
                        -.01,0.39,-0.1,0.12,     /* ok */
                        0.48,-0.30,-0.095,-.02};

static float dat96s[16]={0.07,0.23,-0.37,-.05,  /*Same for 3D configuration */
                        0.35,-0.11,.21,-0.29,   /* LaSilla96, K band short  */ 
                        0.08,-.40,-.02,0.31,    /* not fully ok, need some  */
                        -.3,-0.11,-0.00,.07};   /* lab data. Now copy of KL */

static float dat96l[16]={0.07,0.23,-0.37,-.05,  /*Same for 3D configuration */
                        0.35,-0.11,.21,-0.19,   /* LaSilla96, K band long*/ 
                        0.08,-.40,-.02,0.31,    /* ok */
                        -.3,-0.11,-0.00,.07};

static float dat96h[16]={0.39,0.48,-0.25,-.03,  /*Same for 3D configuration */
                        0.2,-0.21,.09,-0.42,   /* LaSilla96, H band */ 
                        -.01,0.33,-0.15,0.1,    /* ok */
                        0.3,-0.50,-0.3,-.25};

static float dat96k[16]={0.06,0.0,-0.25,-.05,  /*Same for 3D configuration */
                        0.20,-0.19,.09,-0.44,   /* LaSilla96, K band */ 
                        -.03,0.36,-0.1,0.12,
                        0.28,-0.50,-0.34,-.32};

static float dat97k[16]={0.06,0.0,-0.25,-.05,   /*Same for 3D configuration */
			0.20,-0.19,.09,-0.44,   /* Calar Alto 97, K band */
			-.03,0.36,-0.1,0.12,
			0.28,-0.50,-0.34,-.32};

static float dat97hh[16]={0.00,0.12,0.25,-0.42,  /*Same for 3D configuration */
			 0.05,0.35,0.01,0.29,    /*  Calar Alto 97, HH band  */
			 -0.33,0.19,0.48,-0.07,
			 0.01,0.36,-0.44,-0.46};

int Fits::makeCube3d(int year, char band, float scale)
{
	int k, kpix = 0, space, column, kpixshift[16], icol1, icol2;
//	int icol3, icol4;
	int i, iscale, kpixread = 0, n, kslit;
	Fits cube;
	char inputfile[256];
	FILE *fp;
	float datasave[16], datas[16], col1, col2, col3, col4;
	double fkey[2];

	if ((naxis[0] != 2) || (naxis[1] != 256) || (naxis[2] != 600))
		return fits_error("Fits::makeCube3d: must be array of size 256x600");
	cube.create(16,16,600);
	cube = 0.0;
	
	if (!setType(R4)) return FALSE;
	iscale = (int)(scale*1000.+0.5);
	if (iscale != 300 && iscale != 500 && year <= 97  ) {scale=0.5;}
	if (iscale != 250 && iscale != 400 && year == 971 ) {scale=0.4;}
	if (iscale != 250 && iscale != 400 && year == 972 ) {scale=0.4;}
	if (iscale != 250 && iscale != 400 && year == 981 ) {scale=0.4;}
	if (iscale != 250 && iscale != 400 && year == 982 ) {scale=0.4;}

	for (space = 0; space < 600; space++) {

/* Define the # of the first meaningful pixel. */

		if (year == 95) {
			if (band == 'k') {
				kpix=6;
			} else {
				kpix=3;
			}
		} else if (year == 96) {
			if (band == 'l') {
				kpix=4;
			} else if (band == 's') {
				kpix=7;
			} else if (band == 'k') {
				kpix=3;
			} else if (band == 'h') {
				kpix=3;
			}
		} else if (year == 97) {
			if (band == 'k') {
				kpix=7;
			} else if (band == 'g') {
				kpix=5;
			}
		} else if (year == 971 && space == 0) {
			sprintf(inputfile,"3dcubin.971%c", band);
			printf("inputfile = %s\n", inputfile); 
		} else if (year == 972 && space == 0) {
			sprintf(inputfile,"3dcubin.972%c", band);
			printf("inputfile = %s\n", inputfile); 
		} else if (year == 981 && space == 0) {
			sprintf(inputfile,"3dcubin.981%c", band);
			printf("inputfile = %s\n", inputfile); 
		} else if (year == 982 && space == 0) {
			sprintf(inputfile,"3dcubin.982%c", band);
			printf("inputfile = %s\n", inputfile); 
		} else if (year == 993) {
			sprintf(inputfile,"3dcubin.993%c", band);
			printf("inputfile = %s\n", inputfile); 
		} else {
			printf("unknown observing run / band %i %c\n", year, band);
			return FALSE;
		}

		if (year > 97 && space == 0) {
			fp = fopen(inputfile,"r");
			if (!fp) {
				printf("Cannot open input file (%s)!\n", inputfile);
				return FALSE;
			}

         /* Each input file has the format:
            kpixsmall kpixlarge
            datas[1]  kpixshift  datas[1]  kpixshift
                .         .          .         .
                .         .          .         .
                .         .          .         .
            datas[16] kpixshift  datas[16] kpixshift

            The first row is the value of kpix to be used for the small and
            large pixel scale, respectively.
            The subsequent 16 rows represent the value for datas and the
            shift required for kpix which corresponds to the column at the
            end of that slitlet.  These two entries are given for both the
            small and large pixel scales.  Note that *all* slitlets have a
            kpix shift listed, even if the value is zero.
         */
			fscanf(fp,"%d %d",&icol1,&icol2);

			switch(iscale) {
				case 250:
				case 300: 
					kpix = icol1;
					kpixread = kpix;
				break;
				case 400:
				case 500: 
					kpix = icol2;
					kpixread = kpix;
				break;
			}
			for (i=0;i<16;i++) {
				fscanf(fp,"%f %f %f %f",&col1,&col2,&col3,&col4);
				k = (int)(scale*1000.);
/*
datas values now are read into the datasave array for use later (see note
under polynomial interpolation).  L.E. Tacconi-Garman (28.07.98)
*/
				switch(iscale) {
					case 250:
					case 300:
						datasave[i] = col1;
						kpixshift[i] = (int)col2;
					break;
					case 400:
					case 500:
						datasave[i] = col3;
						kpixshift[i] = (int)col4;
					break;
				}
				printf("kpixshift[%2d] = %2d\n",i,kpixshift[i]); 
			}
			n = fclose(fp);
			if (n!=0) {
				printf("Cannot close input file (%s)!\n",inputfile);
				return FALSE;
			}
		} else if (year > 97) {
			kpix = kpixread;
		}

		for (column = 0; column < 256; column++) {
			if (column%16==0 && column != 0) {
				kslit = (int)((float)column/16.);
			} else {
				kslit= -1;
			}
			if (year == 95 && band == 'k') {
            if      (column ==  32) {kpix -=1;} /* Same for the Dec 95/Jan 96 */
            else if (column ==  80) {kpix -=1;} /* K band. */
            else if (column == 112) {kpix -=1;}   
            else if (column == 160) {kpix -=1;}   
            else if (column == 208) {kpix -=1;}
			} else if (year == 95 && band == 'h') {
            if      (column ==  32) {kpix -=1;} /* Same for the Dec 95/Jan 96 */
            else if (column ==  80) {kpix -=1;} /* K band. */
            else if (column == 112) {kpix -=1;}   
            else if (column == 160) {kpix -=1;}   
			} else if (year == 96 && band == 'h') {
            if      (column ==  32) {kpix -=1;} /* Same for the Mar/Apr 96 */
            else if (column ==  80) {kpix -=1;} /* H band. */
            else if (column == 112) {kpix -=1;}   
            else if (column == 160) {kpix -=1;}   
            else if (column == 208) {kpix -=1;}
			} else if (year == 96 && band == 'l') {
            if      (column ==  32) {kpix -=1;} /* Same for the Mar/Aprl 96 */
            else if (column ==  80) {kpix -=1;} /* KL band. */
            else if (column == 112) {kpix -=1;}   
            else if (column == 144) {kpix -=1;}   
            else if (column == 192) {kpix -=1;}
			} else if (year == 96 && band == 's') {
            if      (column ==  32) {kpix -=1;} /* Same for the Mar/Apr 96 */
            else if (column ==  80) {kpix -=1;} /* KS band. */
            else if (column == 112) {kpix -=1;}   
            else if (column == 144) {kpix -=1;}   
            else if (column == 192) {kpix -=1;}
			} else if (year == 96 && band == 'k') {
            if      (column ==  32) {kpix -=1;} /* Same for the Mar/Apr 96 */
            else if (column ==  80) {kpix -=1;} /* K band. */
            else if (column == 112) {kpix -=1;}   
            else if (column == 160) {kpix -=1;}   
            else if (column == 208) {kpix -=1;}
			} else if (year == 97 && band == 'k') {
            if      (column ==  32) {kpix -=1;} /* Same for the Jul/Aug CA97 */
            else if (column ==  80) {kpix -=1;} /* K band. */
            else if (column == 112) {kpix -=1;}    
            else if (column == 160) {kpix -=1;}   
            else if (column == 208) {kpix -=1;}
			} else if (year == 97 && band == 'g') {
            if      (column ==  48) {kpix -=1;} /* Same for the Jul/Aug CA97 */
            else if (column ==  96) {kpix -=1;} /* HH band. */
            else if (column == 128) {kpix -=1;}   
            else if (column == 176) {kpix -=1;}   
            else if (column == 192) {kpix -=1;}
            else if (column == 208) {kpix +=1;}
            else if (column == 224) {kpix -=1;}
			} else if (year > 97) {
				if (kslit >= 0) {
					kpix += kpixshift[kslit];
				}
			}
			if ((kpix-1+space*256 >= 0) && (kpix-1+space*256 < (int)Nelements()))
			cube.r4data[column + space * 256] = r4data[kpix-1+space*256];
			kpix++;
		}
	}

	for (k=0;k<16;k++) { 
		if (year == 95 && band == 'k') {
			datas[k]=dat96[k];
		} else if (year == 95 && band == 'h') {
			datas[k]=dat96[k];
		} else if (year == 96 && band == 'h') {
			datas[k]=dat96h[k];
		} else if (year == 96 && band == 's') {
			datas[k]=dat96s[k];
		} else if (year == 96 && band == 'l') {
			datas[k]=dat96l[k];
		} else if (year == 96 && band == 'k') {
			datas[k]=dat96k[k];
		} else if (year == 97 && band == 'k') {
			datas[k]=dat97k[k];
		} else if (year == 97 && band == 'g') {
			datas[k]=dat97hh[k];
		} else {
			datas[k]=datasave[k];
		}
	}

	for (kpix = 1; kpix <= 16; kpix++) cube.lineshift(kpix, datas[kpix - 1], 1);
	copy(cube);
	fkey[0] = fkey[1] = 7.0;
	SetFloatKeys("CRPIX", fkey, 2);
	fkey[0] = fkey[1] = 0.0;
	SetFloatKeys("CRVAL", fkey, 2);
	fkey[0] = fkey[1] = scale;
	SetFloatKeys("CDELT", fkey, 2);
	SetStringKey("CTYPE1", "relative");
	SetStringKey("CTYPE2", "relative");
	SetStringKey("CUNIT1", "arcsec");
	SetStringKey("CUNIT2", "arcsec");
	SetFloatKey("CDELT3", 0.0005);
	SetFloatKey("CRPIX3", 0.0);
	SetStringKey("CTYPE3", "absolute");
	SetStringKey("CUNIT3", "microns");
	if (band == 'l') {
		SetFloatKey("CRVAL3", 2.15);
	} else if (band == 's') {
		SetFloatKey("CRVAL3", 1.93);
	}
	return(1);
}

int expandCal3d(Fits & a)
{
	int i, j, k;
	float v;
	
	if (!a.setType(R4)) return FALSE;
	for (i = 1; i <= a.Naxis(3); i++) {
		v = 0.0;
		for (j = 1; j <= a.Naxis(1); j++) {
			for (k = 1; k <= a.Naxis(2); k++) {
				v += a.r4data[a.at(j, k, i)];
			}
		}
		v /= (float)(a.Naxis(1) * a.Naxis(2));
		for (j = 1; j <= a.Naxis(1); j++) {
			for (k = 1; k <= a.Naxis(2); k++) {
				a.r4data[a.at(j, k, i)] = v;
			}
		}
	}
	return(1);
}

bool norm3d(Fits & a, const char *fname)
{
	long i, x, y;
	float w[600], v[600];
	FILE *fd;
	
	if ((a.Naxis(0) != 3) || (a.Naxis(1) != 16) || (a.Naxis(2) != 16) || (a.Naxis(3) != 600)) {
		printf("norm3d: must be array of size 16x16x600\n");
		return FALSE;
	}
	if (!a.setType(R4)) return FALSE;
	a.norm();
	if ((fd = fopen(fname, "rb")) == NULL) return FALSE;
	for (i = 0; i < 600; i++) fscanf(fd, "%f %f\n", &w[i], &v[i]);
	fclose(fd);

// datacube for K-long starts at 2.1305, KLEINMANN at 2.1500
	for (i = 100; i < 600; i++) {
		for (x = 1; x <= 16; x++) {
			for (y = 1; y <= 16; y++) {
				a.r4data[a.F_I(x, y, i)] /= v[i-100];
			}
		}
	}
	return TRUE;
}

bool Fits::spec3d(Fits &cube, int x, int y, int r1, int r2, int method) {
    double  value;
    int     i, j, z;
    Fits    _tmp;
    bool    subtract_sky;
    char    key[255];

    subtract_sky = (r2 > r1);
    if (!create(cube.Naxis(3), 1, R4)) {
        return FALSE;
    }
    CopyHeader(cube);
    SetFloatKey("CRVAL1", cube.getCRVAL(3));
    SetFloatKey("CRPIX1", cube.getCRPIX(3));
    SetFloatKey("CDELT1", cube.getCDELT(3));
    if (cube.GetStringKey("CTYPE3", key)) {
        SetStringKey("CTYPE1", key);
    } else {
        DeleteKey("CTYPE1");
    }
    if (cube.GetStringKey("CUNIT3", key)) {
        SetStringKey("CUNIT1", key);
    } else {
        DeleteKey("CUNIT1");
    }
    for (i = 1; i <= MAXNAXIS; i++) {
        for (j = 1; j <= MAXNAXIS; j++) {
            sprintf(key, "CD%i_%i", i, j);
            DeleteKey(key);
        }
    }
    for (i = 2; i <= MAXNAXIS; i++) {
        naxis[i] = 1;
        sprintf(key, "CUNIT%i", i);
        DeleteKey(key);
        sprintf(key, "CTYPE%i", i);
        DeleteKey(key);
        sprintf(key, "CRPIX%i", i);
        DeleteKey(key);
        sprintf(key, "CRVAL%i", i);
        DeleteKey(key);
        sprintf(key, "CDELT%i", i);
        DeleteKey(key);
    }
    DeleteKey("CROTA1");
    DeleteKey("CROTA2");
    sprintf(crtype, "");

    if (subtract_sky && (!_tmp.copy(*this))){
        return FALSE;
    }

    // Calculate spectrum of single pixel at position (x,y)
    if ((r1 == 0) && (r2 <= r1)) {
        if ((x > 0) && (x <= (int)cube.Naxis(1)) && (y > 0) && (y <= (int)cube.Naxis(2))) {
            for (z = 1; z <= cube.Naxis(3); z++) {
                value = cube.ValueAt(cube.F_I(x, y, z));
                if (gsl_finite(value)) {
                    r4data[z-1] = value;
                }
            }
        }
        return TRUE;
    }

    int radius = r2;
    if (r1 > r2) {
        radius = r1;
    }

    // case of median filtering
    float *medmem, *medmem2;
    medmem = (float *)malloc(4*(radius+1)*(radius+1)*sizeof(float));
    medmem2 = (float *)malloc(4*(radius+1)*(radius+1)*sizeof(float));
    int counter, counter2;
    for (z = 1; z <= cube.Naxis(3); z++) {
        counter = counter2 = 0;
        for (i = x - radius; i <= x + radius; i++) {
            for (j = y - radius; j <= y + radius; j++) {
                if ((i > 0) && (i <= (int)cube.Naxis(1)) && (j > 0) && (j <= (int)cube.Naxis(2))) {
                    if ((double)((i-x)*(i-x)) + (double)((j-y)*(j-y)) <= (double)(r1 * r1)) {
                        value = cube.ValueAt(cube.F_I(i, j, z));
                        if (gsl_finite(value)) {
                            medmem[counter] = value;
                            counter++;
                        }
                    } else if (subtract_sky) {
                        if ((double)((i-x)*(i-x)) + (double)((j-y)*(j-y)) <= (double)(r2 * r2)) {
                            value = cube.ValueAt(cube.F_I(i, j, z));
                            if (gsl_finite(value)) {
                                medmem2[counter2] = value;
                                counter2++;
                            }
                        }
                    }
                }
            }
        }
        r4data[z-1] = 0.;
        if (counter > 0) {
            if (method == 0 || method == 1) { // sum, average
                r4data[z-1] = gsl_stats_float_mean(medmem, 1, counter);
            } else { // median
                gsl_sort_float(medmem, 1, counter);
                r4data[z-1] = gsl_stats_float_median_from_sorted_data(medmem, 1, counter);
            }
        }
        if (counter2 > 0) {
            if (method == 0 || method == 1) { // sum, average
                r4data[z-1] -= gsl_stats_float_mean(medmem2, 1, counter2);
            } else { // median
                gsl_sort_float(medmem2, 1, counter2);
                r4data[z-1] -= gsl_stats_float_median_from_sorted_data(medmem2, 1, counter2);
            }
        }
        if (method == 0) { // sum
            r4data[z-1] *= counter;
        }
    }
    free(medmem);
    free(medmem2);

    return TRUE;

    double v1 = 0.0;
    double v2 = 0.0;
    for (i = x - radius; i <= x + radius; i++) {
        for (j = y - radius; j <= y + radius; j++) {
            if ((i > 0) && (i <= (int)cube.Naxis(1)) && (j > 0) && (j <= (int)cube.Naxis(2))) {
                if ((double)((i-x)*(i-x)) + (double)((j-y)*(j-y)) <= (double)(r1 * r1)) {
                    v1++;
                    for (z = 1; z <= cube.Naxis(3); z++) {
                        value = cube.ValueAt(cube.F_I(i, j, z));
                        if (gsl_finite(value)) {
                            r4data[z-1] += value;
                        }
                    }
                } else if (subtract_sky) {
                    if ((double)((i-x)*(i-x)) + (double)((j-y)*(j-y)) <= (double)(r2 * r2)) {
                        v2++;
                        for (z = 1; z <= cube.Naxis(3); z++) {
                            value = cube.ValueAt(cube.F_I(i, j, z));
                            if (gsl_finite(value)) {
                                _tmp.r4data[z-1] += value;
                            }
                        }
                    }
                }
            }
        }
    }
    if (v1 > 0.0) {
        div(v1);
    }
    if (subtract_sky) {
        if (v2 > 0.0) {
            _tmp.div(v2);
        }
        sub(_tmp);
    }
    if (v1 > 0.0) {
        mul(v1);
    }

    return TRUE;
}

/*
bool Fits::spec3d(Fits &cube, int x, int y, int r1, int r2, double reject, int method) {
    double  value;
    int     i, j, z;
    Fits    _tmp;
    bool    subtract_sky;
    char    key[255];

    subtract_sky = (r2 > r1);
    if (!create(cube.Naxis(3), 1, R4)) {
        return FALSE;
    }
    CopyHeader(cube);
    SetFloatKey("CRVAL1", cube.getCRVAL(3));
    SetFloatKey("CRPIX1", cube.getCRPIX(3));
    SetFloatKey("CDELT1", cube.getCDELT(3));
    if (cube.GetStringKey("CTYPE3", key)) {
        SetStringKey("CTYPE1", key);
    } else {
        DeleteKey("CTYPE1");
    }
    if (cube.GetStringKey("CUNIT3", key)) {
        SetStringKey("CUNIT1", key);
    } else {
        DeleteKey("CUNIT1");
    }
    for (i = 1; i <= MAXNAXIS; i++) {
        for (j = 1; j <= MAXNAXIS; j++) {
            sprintf(key, "CD%i_%i", i, j);
            DeleteKey(key);
        }
    }
    for (i = 2; i <= MAXNAXIS; i++) {
        naxis[i] = 1;
        sprintf(key, "CUNIT%i", i);
        DeleteKey(key);
        sprintf(key, "CTYPE%i", i);
        DeleteKey(key);
        sprintf(key, "CRPIX%i", i);
        DeleteKey(key);
        sprintf(key, "CRVAL%i", i);
        DeleteKey(key);
        sprintf(key, "CDELT%i", i);
        DeleteKey(key);
    }
    DeleteKey("CROTA1");
    DeleteKey("CROTA2");
    sprintf(crtype, "");

    if (subtract_sky && (!_tmp.copy(*this))) {
        return FALSE;
    }

    // Calculate spectrum of single pixel at position (x,y)
    if ((r1 == 0) && (r2 <= r1)) {
        if ((x > 0) && (x <= (int)cube.Naxis(1)) && (y > 0) && (y <= (int)cube.Naxis(2))) {
            for (z = 1; z <= cube.Naxis(3); z++) {
                value = cube.ValueAt(cube.F_I(x, y, z));
                if ((value != reject) && gsl_finite(value)) {
                    r4data[z-1] = value;
                }
            }
        }
        return TRUE;
    }

    int radius = r2;
    if (r1 > r2) {
        radius = r1;
    }

// case of median filtering
    float *medmem, *medmem2;
    medmem = (float *)malloc(4*(radius+1)*(radius+1)*sizeof(float));
    medmem2 = (float *)malloc(4*(radius+1)*(radius+1)*sizeof(float));
    int counter, counter2;
    for (z = 1; z <= cube.Naxis(3); z++) {
        counter = counter2 = 0;
        for (i = x - radius; i <= x + radius; i++) {
            for (j = y - radius; j <= y + radius; j++) {
                if ((i > 0) && (i <= (int)cube.Naxis(1)) && (j > 0) && (j <= (int)cube.Naxis(2))) {
                    if ((double)((i-x)*(i-x)) + (double)((j-y)*(j-y)) <= (double)(r1 * r1)) {
                        value = cube.ValueAt(cube.F_I(i, j, z));
                        if ((value != reject) && gsl_finite(value)) {
                            medmem[counter] = value;
                            counter++;
                        }
                    } else if (subtract_sky) {
                        if ((double)((i-x)*(i-x)) + (double)((j-y)*(j-y)) <= (double)(r2 * r2)) {
                            counter2++;
                            value = cube.ValueAt(cube.F_I(i, j, z));
                            if ((value != reject) && gsl_finite(value)) {
                                medmem2[counter2] = value;
                                counter2++;
                            }
                        }
                    }
                }
            }
        }
        r4data[z-1] = 0.;
        if (counter > 0) {
            if (method == 0 || method == 1) { // sum, average
                r4data[z-1] = gsl_stats_float_mean(medmem, 1, counter);
            } else { // median
                gsl_sort_float(medmem, 1, counter);
                r4data[z-1] = gsl_stats_float_median_from_sorted_data(medmem, 1, counter);
            }
        }
        if (counter2 > 0) {
            if (method == 0 || method == 1) { // sum, average
                r4data[z-1] = gsl_stats_float_mean(medmem2, 1, counter2);
            } else { // median
                gsl_sort_float(medmem2, 1, counter2);
                r4data[z-1] -= gsl_stats_float_median_from_sorted_data(medmem2, 1, counter2);
            }
        }
        if (method == 0) { // sum
            r4data[z-1] *= counter;
        }
    }
    free(medmem);
    free(medmem2);

    return TRUE;

    double v1 = 0.0;
    double v2 = 0.0;
    for (i = x - radius; i <= x + radius; i++) {
        for (j = y - radius; j <= y + radius; j++) {
            if ((i > 0) && (i <= (int)cube.Naxis(1)) && (j > 0) && (j <= (int)cube.Naxis(2))) {
                if ((double)((i-x)*(i-x)) + (double)((j-y)*(j-y)) <= (double)(r1 * r1)) {
                    v1++;
                    for (z = 1; z <= cube.Naxis(3); z++) {
                        value = cube.ValueAt(cube.F_I(i, j, z));
                        if ((value != reject) && gsl_finite(value)) {
                            r4data[z-1] += value;
                        }
                    }
                } else if (subtract_sky) {
                    if ((double)((i-x)*(i-x)) + (double)((j-y)*(j-y)) <= (double)(r2 * r2)) {
                        v2++;
                        for (z = 1; z <= cube.Naxis(3); z++) {
                            value = cube.ValueAt(cube.F_I(i, j, z));
                            if ((value != reject) && gsl_finite(value)) {
                                _tmp.r4data[z-1] += value;
                            }
                        }
                    }
                }
            }
        }
    }
    if (v1 > 0.0) {
        div(v1);
    }
    if (subtract_sky) {
        if (v2 > 0.0) {
            _tmp.div(v2);
        }
        _tmp.cblank();
        sub(_tmp);
    }
    if (v1 > 0.0) {
        mul(v1);
    }
    cblank();

    return TRUE;
}
*/

#ifdef WIN
#pragma warning (default: 4305)
#endif
